// ImmutableVector.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "ImmutableVector.h"
#include <iostream>


template <class ImmutableVector>
void DisplayImmutableVector(const char* name, const ImmutableVector& v)
{
	using namespace std;

	cout << name << ".Size() = " << v.Size() 
		 << ", " << name << "[] = { ";
	for (size_t ctr = 0; ctr < v.Size(); ++ctr) {
		cout << v[ctr] << " ";
	}
	cout << "}" << endl;
}


void ImmutableVectorTest1()
{
	// Create an immutable vector with a branching size of four
	ImmutableVector<int, 4> v;

	// Another immutable vector (we will take a copy of v at element 6)
	ImmutableVector<int, 4> aCopyOfV;

	// Push 16 values into the vector
	// (This will create a two level tree)
	// Note that the vector is being assigned to itself. The
	// move constructor insures this is not very expensive, but
	// if a copy was made at any point the copy would remain 
	// unchanged, but continue to share the applicable data with
	// the current version.
	for (int ctr = 0; ctr < 10; ++ctr) {
		v = AppendValue(v, ctr);
		if (ctr == 6) aCopyOfV = v; 
	}

	// Display the contents of the vectors
	DisplayImmutableVector("v", v);
	DisplayImmutableVector("aCopyOfV", aCopyOfV);
}

void ImmutableVectorTest2() 
{
	ImmutableVector<int, 4> v;
	v = AppendValue(v, 1);
	v = AppendValue(v, 2);
	v = AppendValue(v, 3);
    using namespace std;
    cout << "v.Front() = " << v.Front() << endl;
    cout << "v.Back() = " << v.Back() << endl;
	auto v1 = TruncateValue(v);
	auto v2 = SubstituteValueAtIndex(v, 0, 3);
	auto v3 = GenerateFrom(v, [](ImmutableVector<int, 4>::MutableVector& v) {
		v[0] = 4;
		v[1] = 5;
		v[2] = 6;
		v.PushBack(7);
		v.PushBack(8);
	});
	auto v4 = GenerateFrom(v3, [](ImmutableVector<int, 4>::MutableVector& v4) {
		using namespace std;
		cout << "Change v4 by calling PopBack:" << endl;
		cout << "x = v4.PopBack() = " << endl;
		int x = v4.PopBack();
		cout << "x == " << x << endl;
		cout << endl;
	});
	auto v5 = GenerateFrom(ImmutableVector<int, 4>(), [](ImmutableVector<int, 4>::MutableVector& v4) {
		using namespace std;
		cout << "Change v4 by calling PopBack:" << endl;
		cout << "x = v4.PopBack()" << endl;
		v4.PushBack(1);
		int x = v4.PopBack();
		cout << "x == " << x << endl;
		cout << endl;
	});
    auto v6 = GenerateNewImmutableVector<int, 4>([](ImmutableVector<int, 4>::MutableVector& v6) {
        using namespace std;
        cout << "Generate v6 from a lambda" << endl;
        v6.PushBack(10);
        cout << "v6.PushBack(10)" << endl;
        v6.PushBack(20);
        cout << "v6.PushBack(20)" << endl;
        v6.PushBack(30);
        cout << "v6.PushBack(30)" << endl;
        v6[1] = 200;
        cout << "v6[1] = 200" << endl;
        cout << endl;
    });

	// Display the contents of the vectors
	std::cout << "Resulting ImmutableVectors:" << std::endl;
	DisplayImmutableVector("v", v);
	DisplayImmutableVector("v1", v1);
	DisplayImmutableVector("v2", v2);
	DisplayImmutableVector("v3", v3);
    DisplayImmutableVector("v4", v4);
    DisplayImmutableVector("v5", v5);
    DisplayImmutableVector("v6", v6);
}

int _tmain(int /*argc*/, _TCHAR* /*argv*/[])
{
	ImmutableVectorTest1();
	std::cin.get();
	ImmutableVectorTest2();
	std::cin.get();

	return 0;
}
