// Imutable.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <algorithm>
#include <functional>
#include <iostream>
#include <string>
#include <vector>
#include <utility>

using namespace std;
using namespace std::tr1;

class Immutable
{
private:
	// Use a normal double, copying is cheap
	double d_;

	// Use a shared string, since strings can be very large
	std::shared_ptr<std::string const> s_;

public:
	// Default constructor
	Immutable() 
		: d_(0.0),
		s_(std::make_shared<std::string const>(""))
	{}

	// Constructor taking a string
	Immutable(const double d, const string& s) 
		: d_(d),
		s_(std::make_shared<std::string const>(s))
	{}

	// Move constructor
	Immutable(Immutable&& other)
		: s_()
	{
		using std::swap;
		swap(d_, other.d_);
		swap(s_, other.s_);
	}

	// Move assignment operator
	Immutable& operator=(Immutable&& other)
	{
		swap(d_, other.d_);
		swap(s_, other.s_);
		return *this;
	}

	// Getter Functions
	double GetD() const
	{
		// Return a copy since double is small (8 bytes)
		return d_;
	}
	const std::string& GetS() const
	{
		// Return a const reference since string can be very large
		return *s_;
	}

	// "Setter" Functions (always return a new object)
	Immutable SetD(double d) const
	{
		Immutable newObject(*this);
		newObject.d_ = d;
		return newObject;
	}
	Immutable SetS(const std::string& str) const
	{
		Immutable newObject(*this);
		newObject.s_ = std::make_shared<std::string const>(str);
		return newObject;
	}
};

void Foo()
{
	// Create an immutable object
	double d1 = 1.1;
	string s1 = "Hello World";
	Immutable a(d1, s1);

	// Create a copy of the immutable object, share the data
	Immutable b(a);

	// Create a new immutable object
	// by changing an existing immutable object
	// (Note the new object is returned
	string s2 = "Hello Other";
	Immutable c = a.SetS(s2);

	// Display the contents of each object
	cout << "a.GetD() = " << a.GetD() << ", "
		<< "a.GetS() = " << a.GetS()
		<< " [address = " << &(a.GetS()) << "]" << endl;
	cout << "b.GetD() = " << b.GetD() << ", "
		<< "b.GetS() = " << b.GetS()
		<< " [address = " << &(b.GetS()) << "]" << endl;
	cout << "c.GetD() = " << c.GetD() << ", "
		<< "c.GetS() = " << c.GetS()
		<< " [address = " << &(c.GetS()) << "]" << endl;
}

int main(int argc, char* argv[])
{
	Foo();

	cin.get();

	return 0;
}